// Copyright (c) ZeroC, Inc. All rights reserved.

using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace ZeroC.Ice
{
    /// <summary>All mapped structs generated by the Slice compiler implement IStreamableStruct.</summary>
    public interface IStreamableStruct
    {
        /// <summary>Writes the instance to the <see cref="OutputStream"/>.</summary>
        /// <param name="ostr">The <see cref="OutputStream"/> to write to.</param>
        public void IceWrite(OutputStream ostr);
    }

    /// <summary>A delegate that writes a value to an output stream.</summary>
    /// <typeparam name="T">The type of the value to write.</typeparam>
    /// <param name="ostr">The output stream to write to.</param>
    /// <param name="value">The value to write to the stream.</param>
    public delegate void OutputStreamWriter<in T>(OutputStream ostr, T value);

    /// <summary>A delegate that writes a value passed as in-reference to an output stream. This value typically
    /// corresponds to the argument tuple or return value tuple of an operation.</summary>
    /// <typeparam name="T">The type of the value to write (a struct).</typeparam>
    /// <param name="ostr">The output stream to write to.</param>
    /// <param name="value">The value to write to the stream.</param>
    public delegate void OutputStreamValueWriter<T>(OutputStream ostr, in T value) where T : struct;

    /// <summary>A delegate that writes a value passed as in-reference to an output stream. The value contains a
    /// stream parameter that will be sent using the returned stream data writer. This value typically corresponds
    /// to the argument tuple or return value tuple of an operation.</summary>
    /// <typeparam name="T">The type of the value to write (a struct).</typeparam>
    /// <param name="ostr">The output stream to write to.</param>
    /// <param name="value">The value to write to the stream.</param>
    /// <param name="cancel">The cancellation token to cancel the sending of the streamable parameter.</param>
    /// <returns>The stream data writer to write the streamable parameter.</returns>
    public delegate Action<SocketStream> OutputStreamValueWriterWithStreamable<T>(
        OutputStream ostr,
        in T value,
        System.Threading.CancellationToken cancel) where T : struct;

    /// <summary>Writes data into a byte buffer using the Ice encoding.</summary>
    public sealed partial class OutputStream
    {
        /// <summary>Represents a position in the OutputStream's buffer. This position consists of the index of the
        /// segment in the list and the offset into the segment.</summary>
        internal struct Position
        {
            /// <summary>Creates a new position from the segment and offset values.</summary>
            /// <param name="segment">The zero based index of the segment.</param>
            /// <param name="offset">The offset into the segment.</param>
            internal Position(int segment, int offset)
            {
                Segment = segment;
                Offset = offset;
            }

            /// <summary>The zero based index of the segment.</summary>
            internal int Segment;
            /// <summary>The offset into the segment.</summary>
            internal int Offset;
        }

        // Cached OutputStreamWriter static objects used by the generated code

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>bool</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<bool> IceWriterFromBool =
            (ostr, value) => ostr.WriteBool(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>byte</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<byte> IceWriterFromByte =
            (ostr, value) => ostr.WriteByte(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>double</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<double> IceWriterFromDouble =
            (ostr, value) => ostr.WriteDouble(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>float</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<float> IceWriterFromFloat =
            (ostr, value) => ostr.WriteFloat(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>int</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<int> IceWriterFromInt =
            (ostr, value) => ostr.WriteInt(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>long</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<long> IceWriterFromLong =
            (ostr, value) => ostr.WriteLong(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>short</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<short> IceWriterFromShort =
            (ostr, value) => ostr.WriteShort(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>string</c> instances.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<string> IceWriterFromString =
            (ostr, value) => ostr.WriteString(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>uint</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<uint> IceWriterFromUInt =
            (ostr, value) => ostr.WriteUInt(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>ulong</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<ulong> IceWriterFromULong =
            (ostr, value) => ostr.WriteULong(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write <c>ushort</c> values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<ushort> IceWriterFromUShort =
            (ostr, value) => ostr.WriteUShort(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write var int values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<int> IceWriterFromVarInt =
            (ostr, value) => ostr.WriteVarInt(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write var long values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<long> IceWriterFromVarLong =
            (ostr, value) => ostr.WriteVarLong(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write var uint values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<uint> IceWriterFromVarUInt =
            (ostr, value) => ostr.WriteVarUInt(value);

        /// <summary>A <see cref="OutputStreamWriter{T}"/> used to write var ulong values.</summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static readonly OutputStreamWriter<ulong> IceWriterFromVarULong =
            (ostr, value) => ostr.WriteVarULong(value);

        /// <summary>The encoding used when writing to this stream.</summary>
        /// <value>The encoding.</value>
        public Encoding Encoding { get; private set; }

        /// <summary>Determines the current size of the stream. This corresponds to the number of bytes already written
        /// to the stream.</summary>
        /// <value>The current size.</value>
        internal int Size { get; private set; }

        // Gets the position for the next write operation.
        internal Position Tail => _tail;

        private const int DefaultSegmentSize = 256;

        // The number of bytes we use by default when writing a size on a fixed number of byte with the 2.0 encoding.
        private const int DefaultSizeLength = 4;

        private static readonly System.Text.UTF8Encoding _utf8 = new System.Text.UTF8Encoding(false, true);

        private bool InEncapsulation => _startPos != null;

        private bool OldEncoding => Encoding == Encoding.V11;

        // The number of bytes that the stream can hold.
        private int _capacity;

        // Data for the class or exception instance that is currently getting marshaled.
        private InstanceData _current;

        // The segment currently used by write operations, this is usually the last segment of the segment list but it
        // can occasionally be one before last after expanding the list. The tail Position always points to this
        // segment, and the tail offset indicates how much of the segment has been used.
        private ArraySegment<byte> _currentSegment;

        // The current class/exception format, can be either Compact or Sliced.
        private readonly FormatType _format;

        // Map of class instance to instance ID, where the instance IDs start at 2.
        // When writing a top-level encapsulation:
        //  - Instance ID = 0 means null.
        //  - Instance ID = 1 means the instance is encoded inline afterwards.
        //  - Instance ID > 1 means a reference to a previously encoded instance, found in this map.
        private Dictionary<AnyClass, int>? _instanceMap;

        // The encoding used to write the encapsulation header itself.
        private readonly Encoding _mainEncoding;

        // All segments before the tail segment are fully used.
        private readonly IList<ArraySegment<byte>> _segmentList;

        // The start of an encapsulation. When set, we are writing to a top-level encapsulation.
        private readonly Position? _startPos;

        // The position for the next write operation.
        private Position _tail;

        // Map of type ID string to type ID index.
        // When writing into a top-level encapsulation, we assign a type ID index (starting with 1) to each type ID we
        // write, in order.
        private Dictionary<string, int>? _typeIdMap;

        // Write methods for basic types

        /// <summary>Writes a boolean to the stream.</summary>
        /// <param name="v">The boolean to write to the stream.</param>
        public void WriteBool(bool v) => WriteByte(v ? (byte)1 : (byte)0);

        /// <summary>Writes a byte to the stream.</summary>
        /// <param name="v">The byte to write.</param>
        public void WriteByte(byte v)
        {
            Expand(1);
            _currentSegment[_tail.Offset] = v;
            _tail.Offset++;
            Size++;
        }

        /// <summary>Writes a double to the stream.</summary>
        /// <param name="v">The double to write to the stream.</param>
        public void WriteDouble(double v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a float to the stream.</summary>
        /// <param name="v">The float to write to the stream.</param>
        public void WriteFloat(float v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes an int to the stream.</summary>
        /// <param name="v">The int to write to the stream.</param>
        public void WriteInt(int v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a long to the stream.</summary>
        /// <param name="v">The long to write to the stream.</param>
        public void WriteLong(long v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a short to the stream.</summary>
        /// <param name="v">The short to write to the stream.</param>
        public void WriteShort(short v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a size to the stream.</summary>
        /// <param name="v">The size.</param>
        public void WriteSize(int v)
        {
            if (OldEncoding)
            {
                if (v < 255)
                {
                    WriteByte((byte)v);
                }
                else
                {
                    WriteByte(255);
                    WriteInt(v);
                }
            }
            else
            {
                WriteVarULong((ulong)v);
            }
        }

        /// <summary>Writes a string to the stream.</summary>
        /// <param name="v">The string to write to the stream.</param>
        public void WriteString(string v)
        {
            if (v.Length == 0)
            {
                WriteSize(0);
            }
            else if (v.Length <= 100)
            {
                Span<byte> data = stackalloc byte[_utf8.GetMaxByteCount(v.Length)];
                int written = _utf8.GetBytes(v, data);
                WriteSize(written);
                WriteByteSpan(data.Slice(0, written));
            }
            else
            {
                byte[] data = _utf8.GetBytes(v);
                WriteSize(data.Length);
                WriteByteSpan(data.AsSpan());
            }
        }

        /// <summary>Writes a uint to the stream.</summary>
        /// <param name="v">The uint to write to the stream.</param>
        public void WriteUInt(uint v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a ulong to the stream.</summary>
        /// <param name="v">The ulong to write to the stream.</param>
        public void WriteULong(ulong v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes a ushort to the stream.</summary>
        /// <param name="v">The ushort to write to the stream.</param>
        public void WriteUShort(ushort v) => WriteFixedSizeNumeric(v);

        /// <summary>Writes an int to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="v">The int to write to the stream.</param>
        public void WriteVarInt(int v) => WriteVarLong(v);

        /// <summary>Writes a long to stream, using Ice's variable-size integer encoding, with the minimum number of
        /// bytes required by the encoding.</summary>
        /// <param name="v">The long to write to the stream. It must be in the range [-2^61..2^61 - 1].</param>
        public void WriteVarLong(long v)
        {
            int encodedSizeExponent = GetVarLongEncodedSizeExponent(v);
            v <<= 2;
            v |= (uint)encodedSizeExponent;
            Span<byte> data = stackalloc byte[sizeof(long)];
            MemoryMarshal.Write(data, ref v);
            WriteByteSpan(data.Slice(0, 1 << encodedSizeExponent));
        }

        /// <summary>Writes a uint to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="v">The uint to write to the stream.</param>
        public void WriteVarUInt(uint v) => WriteVarULong(v);

        /// <summary>Writes a ulong to stream, using Ice's variable-size integer encoding, with the minimum number of
        /// bytes required by the encoding.</summary>
        /// <param name="v">The ulong to write to the stream. It must be in the range [0..2^62 - 1].</param>
        public void WriteVarULong(ulong v)
        {
            int encodedSizeExponent = GetVarULongEncodedSizeExponent(v);
            v <<= 2;
            v |= (uint)encodedSizeExponent;
            Span<byte> data = stackalloc byte[sizeof(ulong)];
            MemoryMarshal.Write(data, ref v);
            WriteByteSpan(data.Slice(0, 1 << encodedSizeExponent));
        }

        // Write methods for constructed types except class and exception

        /// <summary>Writes an array of fixed-size numeric values, such as int and long, to the stream.</summary>
        /// <param name="v">The array of numeric values.</param>
        public void WriteArray<T>(T[] v) where T : struct => WriteSequence(new ReadOnlySpan<T>(v));

        /// <summary>Writes a dictionary to the stream.</summary>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each value to the stream.</param>
        public void WriteDictionary<TKey, TValue>(
            IReadOnlyDictionary<TKey, TValue> v,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
        {
            WriteSize(v.Count);
            foreach ((TKey key, TValue value) in v)
            {
                keyWriter(this, key);
                valueWriter(this, value);
            }
        }

        /// <summary>Writes a dictionary to the stream. The dictionary's value type is reference type.</summary>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="withBitSequence">When true, encodes entries with a null value using a bit sequence; otherwise,
        /// false.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each non-null value to the stream.</param>
        public void WriteDictionary<TKey, TValue>(
            IReadOnlyDictionary<TKey, TValue?> v,
            bool withBitSequence,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
            where TValue : class
        {
            if (withBitSequence)
            {
                int count = v.Count;
                WriteSize(count);
                BitSequence bitSequence = WriteBitSequence(count);
                int index = 0;
                foreach ((TKey key, TValue? value) in v)
                {
                    keyWriter(this, key);
                    if (value != null)
                    {
                        valueWriter(this, value);
                    }
                    else
                    {
                        bitSequence[index] = false;
                    }
                    index++;
                }
            }
            else
            {
                WriteDictionary((IReadOnlyDictionary<TKey, TValue>)v, keyWriter, valueWriter);
            }
        }

        /// <summary>Writes a dictionary to the stream. The dictionary's value type is a nullable value type.
        /// </summary>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each non-null value to the stream.</param>
        public void WriteDictionary<TKey, TValue>(
            IReadOnlyDictionary<TKey, TValue?> v,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
            where TValue : struct
        {
            int count = v.Count;
            WriteSize(count);
            BitSequence bitSequence = WriteBitSequence(count);
            int index = 0;
            foreach ((TKey key, TValue? value) in v)
            {
                keyWriter(this, key);
                if (value is TValue actualValue)
                {
                    valueWriter(this, actualValue);
                }
                else
                {
                    bitSequence[index] = false;
                }
                index++;
            }
        }

        /// <summary>Writes a nullable proxy to the stream.</summary>
        /// <param name="v">The proxy to write, or null.</param>
        public void WriteNullableProxy<T>(T? v) where T : class, IObjectPrx
        {
            if (v != null)
            {
                v.IceWrite(this);
            }
            else
            {
                if (OldEncoding)
                {
                    Identity.Empty.IceWrite(this);
                }
                else
                {
                    this.Write(ProxyKind20.Null);
                }
            }
        }

        /// <summary>Writes a proxy to the stream.</summary>
        /// <param name="v">The proxy to write. This proxy cannot be null.</param>
        public void WriteProxy<T>(T v) where T : class, IObjectPrx => v.IceWrite(this);

        /// <summary>Writes a sequence of fixed-size numeric values, such as int and long, to the stream.</summary>
        /// <param name="v">The sequence of numeric values represented by a ReadOnlySpan.</param>
        // This method works because (as long as) there is no padding in the memory representation of the
        // ReadOnlySpan.
        public void WriteSequence<T>(ReadOnlySpan<T> v) where T : struct
        {
            WriteSize(v.Length);
            if (!v.IsEmpty)
            {
                WriteByteSpan(MemoryMarshal.AsBytes(v));
            }
        }

        /// <summary>Writes a sequence of fixed-size numeric values, such as int and long, to the stream.</summary>
        /// <param name="v">The sequence of numeric values.</param>
        public void WriteSequence<T>(IEnumerable<T> v) where T : struct
        {
            if (v is T[] vArray)
            {
                WriteArray(vArray);
            }
            else if (v is ImmutableArray<T> vImmutableArray)
            {
                WriteSequence(vImmutableArray.AsSpan());
            }
            else
            {
                WriteSequence(v, (ostr, element) => ostr.WriteFixedSizeNumeric(element));
            }
        }

        /// <summary>Writes a sequence to the stream.</summary>
        /// <param name="v">The sequence to write.</param>
        /// <param name="writer">The delegate that writes each element to the stream.</param>
        public void WriteSequence<T>(IEnumerable<T> v, OutputStreamWriter<T> writer)
        {
            WriteSize(v.Count()); // potentially slow Linq Count()
            foreach (T item in v)
            {
                writer(this, item);
            }
        }

        /// <summary>Writes a sequence to the stream. The elements of the sequence are reference types.</summary>
        /// <param name="v">The sequence to write.</param>
        /// <param name="withBitSequence">True to encode null elements using a bit sequence; otherwise, false.</param>
        /// <param name="writer">The delegate that writes each non-null element to the stream.</param>
        public void WriteSequence<T>(IEnumerable<T?> v, bool withBitSequence, OutputStreamWriter<T> writer)
            where T : class
        {
            if (withBitSequence)
            {
                int count = v.Count(); // potentially slow Linq Count()
                WriteSize(count);
                BitSequence bitSequence = WriteBitSequence(count);
                int index = 0;
                foreach (T? item in v)
                {
                    if (item is T value)
                    {
                        writer(this, value);
                    }
                    else
                    {
                        bitSequence[index] = false;
                    }
                    index++;
                }
            }
            else
            {
                WriteSequence((IEnumerable<T>)v, writer);
            }
        }

        /// <summary>Writes a sequence of nullable values to the stream.</summary>
        /// <param name="v">The sequence to write.</param>
        /// <param name="writer">The delegate that writes each non-null value to the stream.</param>
        public void WriteSequence<T>(IEnumerable<T?> v, OutputStreamWriter<T> writer) where T : struct
        {
            int count = v.Count(); // potentially slow Linq Count()
            WriteSize(count);
            BitSequence bitSequence = WriteBitSequence(count);
            int index = 0;
            foreach (T? item in v)
            {
                if (item is T value)
                {
                    writer(this, value);
                }
                else
                {
                    bitSequence[index] = false;
                }
                index++;
            }
        }

        /// <summary>Writes a mapped Slice struct to the stream.</summary>
        /// <param name="v">The struct instance to write.</param>
        public void WriteStruct<T>(T v) where T : struct, IStreamableStruct => v.IceWrite(this);

        // Write methods for tagged basic types

        /// <summary>Writes a tagged boolean to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The boolean to write to the stream.</param>
        public void WriteTaggedBool(int tag, bool? v)
        {
            if (v is bool value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F1);
                WriteBool(value);
            }
        }

        /// <summary>Writes a tagged byte to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The byte to write to the stream.</param>
        public void WriteTaggedByte(int tag, byte? v)
        {
            if (v is byte value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F1);
                WriteByte(value);
            }
        }

        /// <summary>Writes a tagged double to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The double to write to the stream.</param>
        public void WriteTaggedDouble(int tag, double? v)
        {
            if (v is double value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F8);
                WriteDouble(value);
            }
        }

        /// <summary>Writes a tagged float to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The float to write to the stream.</param>
        public void WriteTaggedFloat(int tag, float? v)
        {
            if (v is float value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F4);
                WriteFloat(value);
            }
        }

        /// <summary>Writes a tagged int to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The int to write to the stream.</param>
        public void WriteTaggedInt(int tag, int? v)
        {
            if (v is int value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F4);
                WriteInt(value);
            }
        }

        /// <summary>Writes a tagged long to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The long to write to the stream.</param>
        public void WriteTaggedLong(int tag, long? v)
        {
            if (v is long value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F8);
                WriteLong(value);
            }
        }

        /// <summary>Writes a tagged size to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The size.</param>
        public void WriteTaggedSize(int tag, int? v)
        {
            if (v is int value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.Size);
                WriteSize(value);
            }
        }

        /// <summary>Writes a tagged short to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The short to write to the stream.</param>
        public void WriteTaggedShort(int tag, short? v)
        {
            if (v is short value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F2);
                WriteShort(value);
            }
        }

        /// <summary>Writes a tagged string to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The string to write to the stream.</param>
        public void WriteTaggedString(int tag, string? v)
        {
            if (v is string value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);
                WriteString(value);
            }
        }

        /// <summary>Writes a tagged uint to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The uint to write to the stream.</param>
        public void WriteTaggedUInt(int tag, uint? v)
        {
            if (v is uint value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F4);
                WriteUInt(value);
            }
        }

        /// <summary>Writes a tagged ulong to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The ulong to write to the stream.</param>
        public void WriteTaggedULong(int tag, ulong? v)
        {
            if (v is ulong value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F8);
                WriteULong(value);
            }
        }

        /// <summary>Writes a tagged ushort to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The ushort to write to the stream.</param>
        public void WriteTaggedUShort(int tag, ushort? v)
        {
            if (v is ushort value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.F2);
                WriteUShort(value);
            }
        }

        /// <summary>Writes a tagged int to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The int to write to the stream.</param>
        public void WriteTaggedVarInt(int tag, int? v) => WriteTaggedVarLong(tag, v);

        /// <summary>Writes a tagged long to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The long to write to the stream.</param>
        public void WriteTaggedVarLong(int tag, long? v)
        {
            if (v is long value)
            {
                var format = (EncodingDefinitions.TagFormat)GetVarLongEncodedSizeExponent(value);
                WriteTaggedParamHeader(tag, format);
                WriteVarLong(value);
            }
        }

        /// <summary>Writes a tagged uint to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The uint to write to the stream.</param>
        public void WriteTaggedVarUInt(int tag, uint? v) => WriteTaggedVarULong(tag, v);

        /// <summary>Writes a tagged ulong to stream, using Ice's variable-size integer encoding.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The ulong to write to the stream.</param>
        public void WriteTaggedVarULong(int tag, ulong? v)
        {
            if (v is ulong value)
            {
                var format = (EncodingDefinitions.TagFormat)GetVarULongEncodedSizeExponent(value);
                WriteTaggedParamHeader(tag, format);
                WriteVarULong(value);
            }
        }

        // Write methods for tagged constructed types except class

        /// <summary>Writes a tagged array of fixed-size numeric values to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The array to write.</param>
        public void WriteTaggedArray<T>(int tag, T[]? v) where T : struct =>
            WriteTaggedSequence(tag, new ReadOnlySpan<T>(v));

        /// <summary>Writes a tagged dictionary with fixed-size entries to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="entrySize">The size of each entry (key + value), in bytes.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each value to the stream.</param>
        public void WriteTaggedDictionary<TKey, TValue>(
            int tag,
            IReadOnlyDictionary<TKey, TValue>? v,
            int entrySize,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
        {
            Debug.Assert(entrySize > 1);
            if (v is IReadOnlyDictionary<TKey, TValue> dict)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);
                int count = dict.Count;
                WriteSize(count == 0 ? 1 : (count * entrySize) + GetSizeLength(count));
                WriteDictionary(dict, keyWriter, valueWriter);
            }
        }

        /// <summary>Writes a tagged dictionary with variable-size elements to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each value to the stream.</param>
        public void WriteTaggedDictionary<TKey, TValue>(
            int tag,
            IReadOnlyDictionary<TKey, TValue>? v,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
        {
            if (v is IReadOnlyDictionary<TKey, TValue> dict)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteDictionary(dict, keyWriter, valueWriter);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged dictionary to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="withBitSequence">When true, encodes entries with a null value using a bit sequence; otherwise,
        /// false.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each value to the stream.</param>
        public void WriteTaggedDictionary<TKey, TValue>(
            int tag,
            IReadOnlyDictionary<TKey, TValue?>? v,
            bool withBitSequence,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
            where TValue : class
        {
            if (v is IReadOnlyDictionary<TKey, TValue?> dict)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteDictionary(dict, withBitSequence, keyWriter, valueWriter);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged dictionary to the stream. The dictionary's value type is a nullable value type.
        /// </summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The dictionary to write.</param>
        /// <param name="keyWriter">The delegate that writes each key to the stream.</param>
        /// <param name="valueWriter">The delegate that writes each non-null value to the stream.</param>
        public void WriteTaggedDictionary<TKey, TValue>(
            int tag,
            IReadOnlyDictionary<TKey, TValue?>? v,
            OutputStreamWriter<TKey> keyWriter,
            OutputStreamWriter<TValue> valueWriter)
            where TKey : notnull
            where TValue : struct
        {
            if (v is IReadOnlyDictionary<TKey, TValue?> dict)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteDictionary(dict, keyWriter, valueWriter);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged proxy to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The proxy to write.</param>
        public void WriteTaggedProxy(int tag, IObjectPrx? v)
        {
            if (v != null)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteProxy(v);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged sequence of fixed-size numeric values to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        public void WriteTaggedSequence<T>(int tag, ReadOnlySpan<T> v) where T : struct
        {
            // A null T[]? or List<T>? is implicitly converted into a default aka null ReadOnlyMemory<T> or
            // ReadOnlySpan<T>. Furthermore, the span of a default ReadOnlyMemory<T> is a default ReadOnlySpan<T>, which
            // is distinct from the span of an empty sequence. This is why the "v != null" below works correctly.
            if (v != null)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);
                int elementSize = Unsafe.SizeOf<T>();
                if (elementSize > 1)
                {
                    // This size is redundant and optimized out by the encoding when elementSize is 1.
                    WriteSize(v.IsEmpty ? 1 : (v.Length * elementSize) + GetSizeLength(v.Length));
                }
                WriteSequence(v);
            }
        }

        /// <summary>Writes a tagged sequence of fixed-size numeric values to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        public void WriteTaggedSequence<T>(int tag, IEnumerable<T>? v) where T : struct
        {
            if (v is IEnumerable<T> value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);

                int elementSize = Unsafe.SizeOf<T>();
                if (elementSize > 1)
                {
                    int count = value.Count(); // potentially slow Linq Count()

                    // First write the size in bytes, so that the reader can skip it. We optimize-out this byte size
                    // when elementSize is 1.
                    WriteSize(count == 0 ? 1 : (count * elementSize) + GetSizeLength(count));
                }
                WriteSequence(value);
            }
        }

        /// <summary>Writes a tagged sequence of variable-size elements to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        /// <param name="writer">The delegate that writes each element to the stream.</param>
        public void WriteTaggedSequence<T>(int tag, IEnumerable<T>? v, OutputStreamWriter<T> writer)
        {
            if (v is IEnumerable<T> value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteSequence(value, writer);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged sequence of fixed-size values to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        /// <param name="elementSize">The fixed size of each element of the sequence, in bytes.</param>
        /// <param name="writer">The delegate that writes each element to the stream.</param>
        public void WriteTaggedSequence<T>(int tag, IEnumerable<T>? v, int elementSize, OutputStreamWriter<T> writer)
            where T : struct
        {
            Debug.Assert(elementSize > 0);
            if (v is IEnumerable<T> value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);

                int count = value.Count(); // potentially slow Linq Count()

                if (elementSize > 1)
                {
                    // First write the size in bytes, so that the reader can skip it. We optimize-out this byte size
                    // when elementSize is 1.
                    WriteSize(count == 0 ? 1 : (count * elementSize) + GetSizeLength(count));
                }
                WriteSize(count);
                foreach (T item in value)
                {
                    writer(this, item);
                }
            }
        }

        /// <summary>Writes a tagged sequence of nullable elements to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        /// <param name="withBitSequence">True to encode null elements using a bit sequence; otherwise, false.</param>
        /// <param name="writer">The delegate that writes each non-null element to the stream.</param>
        public void WriteTaggedSequence<T>(
            int tag,
            IEnumerable<T?>? v,
            bool withBitSequence,
            OutputStreamWriter<T> writer)
            where T : class
        {
            if (v is IEnumerable<T?> value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteSequence(value, withBitSequence, writer);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged sequence of nullable values to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The sequence to write.</param>
        /// <param name="writer">The delegate that writes each non-null element to the stream.</param>
        public void WriteTaggedSequence<T>(int tag, IEnumerable<T?>? v, OutputStreamWriter<T> writer)
            where T : struct
        {
            if (v is IEnumerable<T?> value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                WriteSequence(value, writer);
                EndFixedLengthSize(pos);
            }
        }

        /// <summary>Writes a tagged fixed-size struct to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The struct to write.</param>
        /// <param name="fixedSize">The size of the struct, in bytes.</param>
        public void WriteTaggedStruct<T>(int tag, T? v, int fixedSize) where T : struct, IStreamableStruct
        {
            if (v is T value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.VSize);
                WriteSize(fixedSize);
                value.IceWrite(this);
            }
        }

        /// <summary>Writes a tagged variable-size struct to the stream.</summary>
        /// <param name="tag">The tag.</param>
        /// <param name="v">The struct to write.</param>
        public void WriteTaggedStruct<T>(int tag, T? v) where T : struct, IStreamableStruct
        {
            if (v is T value)
            {
                WriteTaggedParamHeader(tag, EncodingDefinitions.TagFormat.FSize);
                Position pos = StartFixedLengthSize();
                value.IceWrite(this);
                EndFixedLengthSize(pos);
            }
        }

        // Other methods

        /// <summary>Writes a sequence of bits to the stream, and returns this sequence backed by the stream's buffer.
        /// </summary>
        /// <param name="bitSize">The minimum number of bits in the sequence.</param>
        /// <returns>The bit sequence, with all bits set. The actual size of the sequence is a multiple of 8.
        /// </returns>
        public BitSequence WriteBitSequence(int bitSize)
        {
            Debug.Assert(bitSize > 0);
            int size = (bitSize >> 3) + ((bitSize & 0x07) != 0 ? 1 : 0);

            Expand(size);

            int remaining = _currentSegment.Count - _tail.Offset;
            if (size <= remaining)
            {
                // Expand above ensures _tail.Offset is not _currentSegment.Count.
                Span<byte> span = _currentSegment.AsSpan(_tail.Offset, size);
                span.Fill(255);
                _tail.Offset += size;
                Size += size;
                return new BitSequence(span);
            }
            else
            {
                Span<byte> firstSpan = _currentSegment.AsSpan(_tail.Offset);
                firstSpan.Fill(255);
                _currentSegment = _segmentList[++_tail.Segment];
                _tail.Offset = size - remaining;
                Size += size;
                Span<byte> secondSpan = _currentSegment.AsSpan(0, _tail.Offset);
                secondSpan.Fill(255);
                return new BitSequence(firstSpan, secondSpan);
            }
        }

        /// <summary>Computes the minimum number of bytes needed to write a variable-length size with the 2.0 encoding.
        /// </summary>
        /// <remarks>The parameter is a long and not a varulong because sizes and size-like values are usually passed
        /// around as signed integers, even though sizes cannot be negative and are encoded like varulong values.
        /// </remarks>
        internal static int GetSizeLength20(long size)
        {
            Debug.Assert(size >= 0);
            return 1 << GetVarULongEncodedSizeExponent((ulong)size);
        }

        // Constructor for protocol frame header and other non-encapsulated data.
        internal OutputStream(Encoding encoding, IList<ArraySegment<byte>> data, Position startAt = default)
        {
            _mainEncoding = encoding;
            Encoding = encoding;
            Encoding.CheckSupported();
            _format = default; // not used
            _segmentList = data;
            if (_segmentList.Count == 0)
            {
                _currentSegment = ArraySegment<byte>.Empty;
                _capacity = 0;
                Size = 0;
                _tail = default;
            }
            else
            {
                _tail = startAt;
                _currentSegment = _segmentList[_tail.Segment];
                Size = Distance(default);
                _capacity = 0;
                foreach (ArraySegment<byte> segment in _segmentList)
                {
                    _capacity += segment.Count;
                }
            }
        }

        // Constructor that starts an encapsulation.
        internal OutputStream(
            Encoding encoding,
            IList<ArraySegment<byte>> data,
            Position startAt,
            Encoding payloadEncoding,
            FormatType format)
            : this(encoding, data, startAt)
        {
            _format = format;
            _startPos = _tail;
            WriteEncapsulationHeader(payloadEncoding); // with placeholder for size
            if (payloadEncoding == Encoding.V20)
            {
                WriteByte(0); // Placeholder for the compression status
            }
            Encoding = payloadEncoding;
        }

        /// <summary>Computes the amount of data written from the start position to the current position and writes that
        /// size at the start position (as a fixed-length size). The size does not include its own encoded length.
        /// </summary>
        /// <param name="start">The start position.</param>
        /// <param name="sizeLength">The number of bytes used to marshal the size 1, 2 or 4.</param>
        internal void EndFixedLengthSize(Position start, int sizeLength = DefaultSizeLength)
        {
            Debug.Assert(start.Offset >= 0);
            if (OldEncoding)
            {
                RewriteFixedLengthSize11(Distance(start) - 4, start);
            }
            else
            {
                RewriteFixedLengthSize20(Distance(start) - sizeLength, start, sizeLength);
            }
        }

        /// <summary>Completes the current encapsulation (if any) and finishes off the underlying buffer. You should not
        /// write additional data to this output stream or its underlying buffer after calling Finish, however rewriting
        /// previous data (with for example <see cref="EndFixedLengthSize"/>) is fine.</summary>
        internal void Finish()
        {
            if (_startPos is Position startPos)
            {
                Encoding = _mainEncoding;
                int sizeLength = OldEncoding ? 4 : DefaultSizeLength;
                RewriteEncapsulationSize(Distance(startPos) - sizeLength, startPos);
            }

            Debug.Assert(_segmentList.Count - 1 == _tail.Segment);
            _segmentList[^1] = _segmentList[^1].Slice(0, _tail.Offset);
        }

        /// <summary>Writes a size on a fixed number of bytes at the given position of the stream.</summary>
        /// <param name="size">The size to write.</param>
        /// <param name="pos">The position to write to.</param>
        /// <param name="sizeLength">The number of bytes used to encode the size. Can be 1, 2 or 4.</param>
        internal void RewriteFixedLengthSize20(int size, Position pos, int sizeLength = DefaultSizeLength)
        {
            Debug.Assert(pos.Segment < _segmentList.Count);
            Debug.Assert(sizeLength == 1 || sizeLength == 2 || sizeLength == 4);

            Span<byte> data = stackalloc byte[sizeLength];
            data.WriteFixedLengthSize20(size);
            RewriteByteSpan(data, pos);
        }

        /// <summary>Returns the current position and writes placeholder for a fixed-length size value. The
        /// position must be used to rewrite the size later.</summary>
        /// <param name="sizeLength">The number of bytes reserved to write the fixed-length size.</param>
        /// <returns>The position before writing the size.</returns>
        internal Position StartFixedLengthSize(int sizeLength = DefaultSizeLength)
        {
            Position pos = _tail;
            WriteByteSpan(stackalloc byte[OldEncoding ? 4 : sizeLength]); // placeholder for future size
            return pos;
        }

        /// <summary>Writes a span of bytes. The stream capacity is expanded if required, the size and tail position are
        /// increased according to the span length.</summary>
        /// <param name="span">The data to write as a span of bytes.</param>
        internal void WriteByteSpan(ReadOnlySpan<byte> span)
        {
            int length = span.Length;
            Expand(length);
            Size += length;
            int offset = _tail.Offset;
            int remaining = _currentSegment.Count - offset;
            Debug.Assert(remaining > 0); // guaranteed by Expand
            int sz = Math.Min(length, remaining);
            if (length > remaining)
            {
                span.Slice(0, remaining).CopyTo(_currentSegment.AsSpan(offset, sz));
            }
            else
            {
                span.CopyTo(_currentSegment.AsSpan(offset, length));
            }
            _tail.Offset += sz;
            length -= sz;

            if (length > 0)
            {
                _currentSegment = _segmentList[++_tail.Segment];
                if (remaining == 0)
                {
                    span.CopyTo(_currentSegment.AsSpan(0, length));
                }
                else
                {
                    span.Slice(remaining, length).CopyTo(_currentSegment.AsSpan(0, length));
                }
                _tail.Offset = length;
            }
        }

        /// <summary>Writes an empty encapsulation.</summary>
        internal Position WriteEmptyEncapsulation(Encoding encoding)
        {
            encoding.CheckSupported();
            WriteEncapsulationHeader(size: encoding == Encoding.V20 ? 3 : 2, encoding);
            if (encoding == Encoding.V20)
            {
                WriteByte(0); // The compression status, 0 = not-compressed
            }
            return _tail;
        }

        internal void WriteEndpoint(Endpoint endpoint)
        {
            if (endpoint.Protocol == Protocol.Ice1 || OldEncoding)
            {
                this.Write(endpoint.Transport);

                Position startPos = _tail;
                int sizeLength = OldEncoding ? 4 : 2;
                if (endpoint is OpaqueEndpoint opaqueEndpoint)
                {
                    WriteEncapsulationHeader(opaqueEndpoint.ValueEncoding, sizeLength); // with placeholder for size
                    WriteByteSpan(opaqueEndpoint.Value.Span); // WriteByteSpan is not encoding-sensitive
                }
                else
                {
                    // For ice1 and ice2, this corresponds to the protocol's encoding.
                    Encoding payloadEncoding = endpoint.Protocol == Protocol.Ice1 ?
                        Ice1Definitions.Encoding : Encoding.V20;

                    WriteEncapsulationHeader(payloadEncoding, sizeLength); // with placeholder for size
                    Encoding previousEncoding = Encoding;
                    Encoding = payloadEncoding;
                    if (endpoint.Protocol == Protocol.Ice1)
                    {
                        endpoint.WriteOptions(this);
                    }
                    else
                    {
                        WriteString(endpoint.Data.Host);
                        WriteUShort(endpoint.Data.Port);
                        WriteSequence(endpoint.Data.Options, IceWriterFromString);
                    }
                    Encoding = previousEncoding;
                }
                RewriteEncapsulationSize(Distance(startPos) - sizeLength, startPos, sizeLength);
            }
            else
            {
                endpoint.Data.IceWrite(this);
            }
        }

        internal void WriteBinaryContextEntry(int key, ReadOnlySpan<byte> value)
        {
            WriteVarInt(key);
            WriteSize(value.Length);
            WriteByteSpan(value);
        }

        internal void WriteBinaryContextEntry<T>(int key, T value, OutputStreamWriter<T> writer)
        {
            WriteVarInt(key);
            Position pos = StartFixedLengthSize(2); // 2-bytes size place holder
            writer(this, value);
            EndFixedLengthSize(pos, 2);
        }

        private static int Distance(IList<ArraySegment<byte>> data, Position start, Position end)
        {
            // If both the start and end position are in the same array segment just
            // compute the offsets distance.
            if (start.Segment == end.Segment)
            {
                return end.Offset - start.Offset;
            }

            // If start and end position are in different segments we need to accumulate the
            // size from start offset to the end of the start segment, the size of the intermediary
            // segments, and the current offset into the last segment.
            ArraySegment<byte> segment = data[start.Segment];
            int size = segment.Count - start.Offset;
            for (int i = start.Segment + 1; i < end.Segment; ++i)
            {
                checked
                {
                    size += data[i].Count;
                }
            }
            checked
            {
                return size + end.Offset;
            }
        }

        /// <summary>Gets the minimum number of bytes needed to encode a long value with the varlong encoding as an
        /// exponent of 2.</summary>
        /// <param name="value">The value to encode.</param>
        /// <returns>N where 2^N is the number of bytes needed to encode value with Ice's varlong encoding.</returns>
        private static int GetVarLongEncodedSizeExponent(long value)
        {
            if (value < EncodingDefinitions.VarLongMinValue || value > EncodingDefinitions.VarLongMaxValue)
            {
                throw new ArgumentOutOfRangeException($"varlong value `{value}' is out of range", nameof(value));
            }

            return (value << 2) switch
            {
                long b when b >= sbyte.MinValue && b <= sbyte.MaxValue => 0,
                long s when s >= short.MinValue && s <= short.MaxValue => 1,
                long i when i >= int.MinValue && i <= int.MaxValue => 2,
                _ => 3
            };
        }

        /// <summary>Gets the mimimum number of bytes needed to encode a long value with the varulong encoding as an
        /// exponent of 2.</summary>
        /// <param name="value">The value to encode.</param>
        /// <returns>N where 2^N is the number of bytes needed to encode value with varulong encoding.</returns>
        private static int GetVarULongEncodedSizeExponent(ulong value)
        {
            if (value > EncodingDefinitions.VarULongMaxValue)
            {
                throw new ArgumentOutOfRangeException($"varulong value `{value}' is out of range", nameof(value));
            }

            return (value << 2) switch
            {
                ulong b when b <= byte.MaxValue => 0,
                ulong s when s <= ushort.MaxValue => 1,
                ulong i when i <= uint.MaxValue => 2,
                _ => 3
            };
        }

        /// <summary>Returns the distance in bytes from start position to the current position.</summary>
        /// <param name="start">The start position from where to calculate distance to current position.</param>
        /// <returns>The distance in bytes from the current position to the start position.</returns>
        private int Distance(Position start)
        {
            Debug.Assert(_tail.Segment > start.Segment ||
                        (_tail.Segment == start.Segment && _tail.Offset >= start.Offset));

            return Distance(_segmentList, start, _tail);
        }

        /// <summary>Expands the stream to make room for more data. If the bytes remaining in the stream are not enough
        /// to hold the given number of bytes, allocates a new byte array. The caller should then consume the new bytes
        /// immediately; calling Expand repeatedly is not supported.</summary>
        /// <param name="n">The number of bytes to accommodate in the stream.</param>
        private void Expand(int n)
        {
            Debug.Assert(n > 0);
            int remaining = _capacity - Size;
            if (n > remaining)
            {
                int size = Math.Max(DefaultSegmentSize, _currentSegment.Count * 2);
                size = Math.Max(n - remaining, size);
                byte[] buffer = new byte[size];
                _segmentList.Add(buffer);
                if (_segmentList.Count == 1) // First Expand for a new OutputStream constructed with no buffer.
                {
                    Debug.Assert(_currentSegment.Count == 0);
                    _currentSegment = buffer;
                }
                else if (remaining == 0)
                {
                    // Patch _tail to point to the first byte in the new buffer.
                    Debug.Assert(_tail.Offset == _currentSegment.Count);
                    _currentSegment = buffer;
                    _tail.Segment++;
                    _tail.Offset = 0;
                }
                _capacity += buffer.Length;
            }

            // Once Expand returns, _tail points to a writeable byte.
            Debug.Assert(_tail.Offset < _currentSegment.Count);
        }

        /// <summary>Computes the minimum number of bytes needed to write a variable-length size with the current
        /// encoding.</summary>
        /// <param name="size">The size.</param>
        /// <returns>The minimum number of bytes.</returns>
        private int GetSizeLength(int size) => OldEncoding ? (size < 255 ? 1 : 5) : GetSizeLength20(size);

        /// <summary>Writes a byte at a given position of the stream.</summary>
        /// <param name="v">The byte value to write.</param>
        /// <param name="pos">The position to write to.</param>
        private void RewriteByte(byte v, Position pos)
        {
            ArraySegment<byte> segment = _segmentList[pos.Segment];
            if (pos.Offset < segment.Count)
            {
                segment[pos.Offset] = v;
            }
            else
            {
                // (segN, segN.Count) points to the same byte as (segN + 1, 0)
                Debug.Assert(pos.Offset == segment.Count);
                segment = _segmentList[pos.Segment + 1];
                segment[0] = v;
            }
        }

        /// <summary>Rewrites an encapsulation size on a fixed number of bytes at the given position of the stream.
        /// </summary>
        /// <param name="size">The number of bytes in the encapsulation, without taking into account the bytes for the
        /// size itself.</param>
        /// <param name="pos">The position to write to.</param>
        /// <param name="sizeLength">The number of bytes used to encode the size with the 2.0 encoding. Can be 1, 2 or
        /// 4.</param>
        private void RewriteEncapsulationSize(int size, Position pos, int sizeLength = DefaultSizeLength)
        {
            if (OldEncoding)
            {
                // With the 1.1 encoding, sizeLength is always 4 bytes and the encoded size includes this sizeLength.
                RewriteFixedLengthSize11(size + 4, pos);
            }
            else
            {
                RewriteFixedLengthSize20(size, pos, sizeLength);
            }
        }

        /// <summary>Writes a size on 4 bytes at the given position of the stream.</summary>
        /// <param name="size">The size to write.</param>
        /// <param name="pos">The position to write to.</param>
        internal void RewriteFixedLengthSize11(int size, Position pos)
        {
            Debug.Assert(pos.Segment < _segmentList.Count);

            Span<byte> data = stackalloc byte[4];
            MemoryMarshal.Write(data, ref size);
            RewriteByteSpan(data, pos);
        }

        private void RewriteByteSpan(Span<byte> data, Position pos)
        {
            ArraySegment<byte> segment = _segmentList[pos.Segment];
            int remaining = Math.Min(data.Length, segment.Count - pos.Offset);
            if (remaining > 0)
            {
                data.Slice(0, remaining).CopyTo(segment.AsSpan(pos.Offset, remaining));
            }

            if (remaining < data.Length)
            {
                segment = _segmentList[pos.Segment + 1];
                data[remaining..].CopyTo(segment.AsSpan(0, data.Length - remaining));
            }
        }

        /// <summary>Writes an encapsulation header.</summary>
        /// <param name="size">The size of the encapsulation, in bytes. This size does not include the length of the
        /// encoded size itself.</param>
        /// <param name="encoding">The encoding of the new encapsulation.</param>
        internal void WriteEncapsulationHeader(int size, Encoding encoding)
        {
            if (OldEncoding)
            {
                WriteInt(size + 4); // the size length is included in the encoded size with the 1.1 encoding.
            }
            else
            {
                WriteSize(size);
            }

            WriteByte(encoding.Major);
            WriteByte(encoding.Minor);
        }

        /// <summary>Writes an encapsulation header with a placeholder size.</summary>
        /// <param name="encoding">The encoding of the new encapsulation.</param>
        /// <param name="sizeLength">The number of bytes used to encode the size, used only with the 2.0 encoding. Can
        /// be 1, 2 or 4.</param>
        private void WriteEncapsulationHeader(Encoding encoding, int sizeLength = DefaultSizeLength)
        {
            Debug.Assert(sizeLength == 1 || sizeLength == 2 || sizeLength == 4);
            WriteByteSpan(stackalloc byte[OldEncoding ? 4 : sizeLength]); // placeholder for future size
            WriteByte(encoding.Major);
            WriteByte(encoding.Minor);
        }

        /// <summary>Writes a fixed-size numeric value to the stream.</summary>
        /// <param name="v">The numeric value to write to the stream.</param>
        private void WriteFixedSizeNumeric<T>(T v) where T : struct
        {
            int elementSize = Unsafe.SizeOf<T>();
            Span<byte> data = stackalloc byte[elementSize];
            MemoryMarshal.Write(data, ref v);
            WriteByteSpan(data);
        }

        /// <summary>Writes the header for a tagged parameter or data member.</summary>
        /// <param name="tag">The numeric tag associated with the parameter or data member.</param>
        /// <param name="format">The tag format.</param>
        private void WriteTaggedParamHeader(int tag, EncodingDefinitions.TagFormat format)
        {
            Debug.Assert(InEncapsulation);
            Debug.Assert(format != EncodingDefinitions.TagFormat.VInt); // VInt cannot be marshaled

            int v = (int)format;
            if (tag < 30)
            {
                v |= tag << 3;
                WriteByte((byte)v);
            }
            else
            {
                v |= 0x0F0; // tag = 30
                WriteByte((byte)v);
                WriteSize(tag);
            }
            if (_current.InstanceType != InstanceType.None)
            {
                _current.SliceFlags |= EncodingDefinitions.SliceFlags.HasTaggedMembers;
            }
        }
    }
}
